//! Simplest sync [Monad].
//!
//! For async, see [`super::future`].

use crate::func::*;

#[derive(SharedFunctor)]
pub struct SoloInstance;

impl WeakFunctor for SoloInstance {
    type F<'a, A: 'a> = A;
}

impl<'a> Functor<'a> for SoloInstance {
    fn fmap<A: 'a, B: 'a>(f: impl 'a + FnOnce(A) -> B, fa: Self::Fa<A>) -> Self::Fa<B> {
        f(fa)
    }

    fn replace<A: 'a, B: 'a>(fa: Self::Fa<A>, b: B) -> Self::Fa<B> {
        drop(fa);
        b
    }

    fn void<A: 'a>(fa: Self::Fa<A>) -> Self::Fa<()> {
        drop(fa);
    }
}

impl<'a> Pure<'a> for SoloInstance {
    fn pure<A: 'a>(a: A) -> Self::Fa<A> {
        a
    }
}

impl<'a> ApplicativeSeq<'a> for SoloInstance {
    fn seq<A: 'a, B: 'a>(ff: Self::Fa<impl 'a + FnOnce(A) -> B>, fa: Self::Fa<A>) -> Self::Fa<B> {
        ff(fa)
    }
}

impl<'a> ApplicativeLA2<'a> for SoloInstance {
    fn la2<A: 'a, B: 'a, C: 'a>(
        f: impl 'a + FnOnce(A, B) -> C,
        fa: Self::Fa<A>,
        fb: Self::Fa<B>,
    ) -> Self::Fa<C> {
        f(fa, fb)
    }
}

impl<'a> ApplicativeTuple<'a> for SoloInstance {
    fn tuple<A: 'a, B: 'a>((fa, fb): (Self::Fa<A>, Self::Fa<B>)) -> Self::Fa<(A, B)> {
        (fa, fb)
    }
}

impl<'a> ApplicativeSelect<'a> for SoloInstance {}

impl<'a> Applicative<'a> for SoloInstance {
    fn discard_first<A: 'a, B: 'a>(fa: Self::Fa<A>, fb: Self::Fa<B>) -> Self::Fa<B> {
        drop(fa);
        fb
    }

    fn discard_second<A: 'a, B: 'a>(fa: Self::Fa<A>, fb: Self::Fa<B>) -> Self::Fa<A> {
        drop(fb);
        fa
    }
}

impl<'a> Monad<'a> for SoloInstance {
    fn bind<A: 'a, B: 'a>(fa: Self::Fa<A>, f: impl 'a + FnOnce(A) -> Self::Fa<B>) -> Self::Fa<B> {
        f(fa)
    }

    fn iterate<B: 'a>(mut f: impl Iterative<'a, T = Self, B = B>) -> Self::Fa<B> {
        loop {
            match f.next() {
                ControlFlow::Continue(next_f) => f = next_f,
                ControlFlow::Break(b) => return b,
            }
        }
    }

    fn join<A: 'a>(ffa: Self::Fa<Self::Fa<A>>) -> Self::Fa<A> {
        ffa
    }
}

impl<'a> LocalFunctor<'a> for SoloInstance {
    fn unstuff<A: 'a, B: 'a>(state: Self::Fa<ControlFlow<B, A>>) -> ControlFlow<Self::Fa<B>, A> {
        state
    }

    fn stuff<A: 'a, T: Pure<'a>>(fa: Self::Fa<T::Fa<A>>) -> T::Fa<Self::Fa<A>> {
        fa
    }
}

impl<'a> Fail<'a, std::convert::Infallible> for SoloInstance {
    fn fail<A: 'a>(e: std::convert::Infallible) -> Self::Fa<A> {
        match e {}
    }
}
